
# This file was generated by 'versioneer.py' from
# revision-control system data, or from the parent directory name of an
# unpacked source archive. Distribution tarballs contain a pre-generated copy
# of this file.

version_version = '0.3.3'
version_full = '26fe50ecda001fc3ea59ea03fd7d0274095e88b6'
def get_versions(default={}, verbose=False):
    return {'version': version_version, 'full': version_full}

